use std::ops::{Range, Index, IndexMut};

// Fixed point: 0xFFFF_FFFF.FFFF_FFFF
#[derive(Clone, Copy)]
pub struct Sample(pub i64);

impl Sample {
    pub fn to_i16_sample(&self) -> i16 {
        // interpret self as fixed-point
        if self.0 > i32::MAX as i64 { return i16::MAX }
        if self.0 < i32::MIN as i64 { return i16::MIN }
        let low_part = self.0 as i32;
        return (low_part >> 16) as i16
    }
}

impl Default for Sample {
    fn default() -> Self { BLANK_SAMPLE }
}

const BLANK_SAMPLE: Sample = Sample(0);

#[derive(Clone)]
pub struct Channel {
    formal: Range<i32>,

    minus: Vec<Sample>,
    plus: Vec<Sample>,
}

pub struct Bounds {
    pub formal: Range<i32>,
    pub informal: Range<i32>,
}

impl Channel {
    pub fn new(formal: Range<i32>) -> Self {
        assert!(formal.end >= formal.start);
        Channel {
            formal: formal,
            minus: vec![],
            plus: vec![],
        }
    }

    pub fn formal(&self) -> Range<i32> {
        self.formal.clone()
    }

    pub fn bounds(&self) -> Bounds {
        let formal = self.formal.clone();
        let informal_start = formal.start - self.minus.len() as i32;
        let informal_len_0 = self.minus.len() + self.plus.len();
        let informal_len_1 = (formal.end - informal_start) as usize; 
        let informal_len = informal_len_0.max(informal_len_1);

        return Bounds {
            formal,
            informal: informal_start..(informal_start + informal_len as i32)
        }
    }

    pub(crate) fn blank() -> Channel {
        return Channel::new(0..0);
    }

    pub fn slice(&self, start: i32, length: i32) -> Channel {
        if start == 0 && length == 0 {
            return self.clone()
        }
        if start != 0 && length == 0 {
            let formal_start = self.formal.start + start;
            let mut formal_end = self.formal.start + self.formal.end;
            if formal_end < formal_start { formal_end = formal_start; }

            let mut chan = Channel::new(formal_start..formal_end);
            for i in 0..self.plus.len() {
                chan[self.formal.start + i as i32] = self[self.formal.start + i as i32]
            }
            return chan
        }

        let formal_start = self.formal.start + start;
        let formal_end = self.formal.start + start + length;
        let mut chan = Channel::new(formal_start..formal_end);
        for i in formal_start..formal_end {
            chan[i] = self[i];
        }
        return chan
    }
} 

impl Index<i32> for Channel {
    type Output = Sample;

    fn index(&self, index: i32) -> &Self::Output {
        let adj_idx = index - self.formal.start;
        if adj_idx < 0 {
            let neg_adj_idx = (-adj_idx) as usize - 1;
            self.minus.get(neg_adj_idx).unwrap_or(&BLANK_SAMPLE)
        } else {
            let pos_adj_idx = adj_idx as usize;
            self.plus.get(pos_adj_idx).unwrap_or(&BLANK_SAMPLE)
        }
    }
}

impl IndexMut<i32> for Channel {
    fn index_mut(&mut self, index: i32) -> &mut Self::Output {
        let adj_idx = index - self.formal.start;
        if adj_idx < 0 {
            let neg_adj_idx = (-adj_idx) as usize - 1;
            if self.minus.len() < neg_adj_idx + 1 {
                self.minus.reserve(neg_adj_idx + 1);
                while self.minus.len() < neg_adj_idx + 1 {
                    self.minus.push(BLANK_SAMPLE);
                }
            }
            return &mut self.minus[neg_adj_idx]
        } else {
            let pos_adj_idx = adj_idx as usize;
            if self.plus.len() < pos_adj_idx + 1{
                self.plus.reserve(pos_adj_idx + 1);
                while self.plus.len() < pos_adj_idx + 1 {
                    self.plus.push(BLANK_SAMPLE);
                }
            }
            return &mut self.plus[pos_adj_idx]
        }
    }
}