use std::{collections::HashMap, cell::Cell};

use crate::{audio::Channel, error::{Lansky, LanskyError}, vm::{Procedure, Block}, rack::Rack};

pub struct Device {
    proto: Proto,
    implementation: Implementation,
}

#[derive(Clone)]
pub struct Proto {  
    pub name: String,
    pub inputs: Vec<String>,
    pub outputs: Vec<String>,
}


enum Implementation {
    Native(fn(Vec<Channel>) -> Lansky<Vec<Channel>>),
    Hosted(Procedure),
}

impl Device {
    pub fn native(
        name: String, 
        inputs: Vec<String>, 
        outputs: Vec<String>, 
        native: fn(Vec<Channel>) -> Lansky<Vec<Channel>>
    ) -> Lansky<Self> {
        let proto = Proto { name, inputs, outputs };
        Self::new(proto, Implementation::Native(native))
    }

    pub fn hosted(
        name: String, 
        inputs: Vec<String>, 
        outputs: Vec<String>, 
        blocks: Vec<Block>,
    ) -> Lansky<Self> {
        let proto = Proto { name, inputs, outputs };
        Self::new(proto.clone(), Implementation::Hosted(Procedure::new(proto, blocks)))
    }

    fn new(proto: Proto, implementation: Implementation) -> Lansky<Self> {
        for i0 in 0..proto.inputs.len() {
            for i1 in i0+1..proto.inputs.len() {
                if proto.inputs[i0] == proto.inputs[i1] { 
                    return Err(LanskyError::DuplicateInput { name: proto.inputs[i0].clone(), i0, i1 }) 
                }
            }
        }

        for i0 in 0..proto.outputs.len() {
            for i1 in i0+1..proto.outputs.len() {
                if proto.outputs[i0] == proto.outputs[i1] { 
                    return Err(LanskyError::DuplicateOutput { name: proto.outputs[i0].clone(), i0, i1 }) 
                }
            }
        }

        // TODO: Validate the code in some way?
        // Not needed for native devices

        Ok(Device { proto, implementation })
    }

    pub(crate) fn analyze(&self, rack: &Rack) -> Lansky<()> {
        match &self.implementation {
            Implementation::Native(_) => {}
            Implementation::Hosted(h) => {
                h.analyze(rack)?;
            }
        }
        Ok(())
    }

    pub fn get_proto(&self) -> &Proto {
        &self.proto
    }

    pub(crate) fn invoke_named(&self, rack: &Rack, mut inputs: HashMap<String, Channel>) -> Lansky<HashMap<String, Channel>> {
        let expected = self.proto.inputs.len();
        let got = inputs.len();
        if expected != got { return Err(LanskyError::WrongNInputs { expected, got }) }

        let mut inputs_2 = Vec::with_capacity(self.proto.inputs.len());
        for i in self.proto.inputs.iter() {
            let Some(channel_in) = inputs.remove(i) else {
                return Err(LanskyError::MissingInput { name: i.clone() })
            };
            inputs_2.push(channel_in)
        }
        for i in inputs.keys() {
            if !self.proto.inputs.contains(i) {
                return Err(LanskyError::UnknownInput { name: i.clone() })
            }
        }

        let outputs = self.invoke_positional(rack, inputs_2)?;

        let mut outputs_2 = HashMap::new();
        for (k, v) in self.proto.outputs.iter().zip(outputs.into_iter()) {
            outputs_2.insert(k.clone(), v);
        }
        Ok(outputs_2)
    }

    pub(crate) fn invoke_positional(&self, rack: &Rack, inputs: Vec<Channel>) -> Lansky<Vec<Channel>> {
        let expected = self.proto.inputs.len();
        let got = inputs.len();
        if expected != got { return Err(LanskyError::WrongNInputs { expected, got }) }
        let outputs = match &self.implementation {
            Implementation::Native(f) => f(inputs)?,
            Implementation::Hosted(procedure) => procedure.invoke(rack, inputs)?,
        };

        let expected = self.proto.outputs.len();
        let got = outputs.len();
        if expected != got { return Err(LanskyError::WrongNOutputs { expected, got }) }
        Ok(outputs)
    }
}