use std::collections::HashMap;

use crate::{device::{Device, Proto}, error::{Lansky, LanskyError}, standard_devices, audio::Channel};

pub struct Rack {
    devices: HashMap<String, Device>
}

impl Rack {
    pub fn new() -> Self {
        let mut r = Rack {
            devices: HashMap::new()
        };
        standard_devices::augment(&mut r);
        r
    }

    pub fn analyze_all(&self) -> Lansky<()> {
        for d in self.devices.values() {
            d.analyze(self)?;
        }
        Ok(())
    }

    pub fn add_device(&mut self, device: Device) -> Lansky<()> {
        if self.devices.contains_key(&device.get_proto().name) {
            return Err(LanskyError::DuplicateDevice { name: device.get_proto().name.clone() })
        }

        let name = device.get_proto().name.clone();
        self.devices.insert(name, device);

        Ok(())
    }

    fn get_device(&self, name: &str) -> Lansky<&Device> {
        let Some(device) = self.devices.get(name) else {
            return Err(LanskyError::UnknownDevice { name: name.to_string() })
        };

        Ok(device)
    }

    pub fn invoke_named(&self, name: &str, inputs: HashMap<String, Channel>) -> Lansky<HashMap<String, Channel>> {
        self.get_device(name)?.invoke_named(self, inputs)
    }

    pub fn invoke_positional(&self, name: &str, inputs: Vec<Channel>) -> Lansky<Vec<Channel>> {
        self.get_device(name)?.invoke_positional(self, inputs)
    }

    pub fn get_proto(&self, name: &str) -> Lansky<&Proto> {
        Ok(self.get_device(name)?.get_proto())
    }

}